//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// @author Brian Ricks
//

#include "RadioPacketGenerator.h"

#include "inet/linklayer/ieee80211/mac/Ieee80211Frame_m.h"
#include "inet/physicallayer/ieee80211/packetlevel/Ieee80211ControlInfo_m.h"

#include <omnetpp.h>

Define_Module(RadioPacketGenerator);

RadioPacketGenerator::RadioPacketGenerator() {
    isScheduled = false;
    // Initial state is undefined as we do not know what the actual radio state is until we receive
    // the first signal from the radio.
    txState = inet::physicallayer::IRadio::TransmissionState::TRANSMISSION_STATE_UNDEFINED;
    txRadioMode = inet::physicallayer::IRadio::RadioMode::RADIO_MODE_OFF;
}

RadioPacketGenerator::~RadioPacketGenerator() {
    // TODO Auto-generated destructor stub
}

void RadioPacketGenerator::initialize(int stage) {
    if (stage == 0) {
        generatedPktLength = par("packetLength");
        sendDelay = par("packetSendDelay");

        // Subscribe to the radio's state change signal.
        cModule *radioModule = gate("radioOut")->getNextGate()->getOwnerModule();  // grab a pointer to the radio (for verification)
        // Get the container module.  This will be the first module to receive the radio's signals as these signals
        // propagate up the module hierarchy.
        radioModule->subscribe(inet::physicallayer::IRadio::radioModeChangedSignal, this);
        radioModule->subscribe(inet::physicallayer::IRadio::transmissionStateChangedSignal, this);

        scheduleMsg = new cMessage("JammerScheduleMsg");

        txRadio = check_and_cast<inet::physicallayer::IRadio *>(radioModule);
    }
    else if (stage == inet::INITSTAGE_LINK_LAYER) {
        isScheduled = true;

        txRadio->setRadioMode(inet::physicallayer::IRadio::RADIO_MODE_TRANSMITTER);

        scheduleAt(simTime() + par("startSendDelay"), scheduleMsg);
    }
}

// Processes self messages as they are received (think of these as timers)
// Any other messages received (such as those received by the radio when it isn't transmitting)
// are discarded.
void RadioPacketGenerator::handleMessage(cMessage *msg) {
    if (msg == scheduleMsg) {
        attemptNextPacketTransmission();
    }
    else {
        // nothing for us, delete
        EV << "[JAMMER]: Received message from Radio, discarding...\n";

        delete msg;
    }
}

// Handler for notifications that we are subscribed to (radio state changes)
void RadioPacketGenerator::receiveSignal(cComponent *source, simsignal_t signalID, long val, cObject *details) {
    Enter_Method_Silent(); // required as this method is callable from other modules

    if (signalID == inet::physicallayer::IRadio::radioModeChangedSignal) {
        // Radio mode change.  Note this change
        EV << "[JAMMER]: Radio mode change.\n";

        txRadioMode = (inet::physicallayer::IRadio::RadioMode) val;

        if (txRadioMode == inet::physicallayer::IRadio::RadioMode::RADIO_MODE_TRANSMITTER && !isScheduled) {

            isScheduled = true;
            attemptNextPacketTransmission();
        }
    }
    else if (signalID == inet::physicallayer::IRadio::transmissionStateChangedSignal) {
        // update the transmission state
        EV << "[JAMMER]: Radio transmission state change.\n";

        txState = (inet::physicallayer::IRadio::TransmissionState) val;

        if (txState == inet::physicallayer::IRadio::TransmissionState::TRANSMISSION_STATE_IDLE && !isScheduled) {

            isScheduled = true;
            attemptNextPacketTransmission();
        }
    }

    return;
}

void RadioPacketGenerator::attemptNextPacketTransmission() {
    // Time to send a message!!

    if (txRadioMode != inet::physicallayer::IRadio::RADIO_MODE_TRANSMITTER) {
        txRadio->setRadioMode(inet::physicallayer::IRadio::RADIO_MODE_TRANSMITTER);

        EV << "[JAMMER]: Radio mode not applicable for transmission.  Changing mode...\n";

        isScheduled = false;

        return;
    }

    if (txState != inet::physicallayer::IRadio::TransmissionState::TRANSMISSION_STATE_IDLE) {
        // wait, what?  Well we can't send now!
        // This will happen if the time it takes to transmit a packet is greater than the send delay.
        // Wait and send when the state becomes desirable.  We assume that once this happens, a signal
        // will be sent alerting us as such.

        EV << "[JAMMER]: Radio transmitter busy.  Waiting for idle state...\n";

        isScheduled = false;

        return;
     }

     sendPacket();
     //schedule the next one
     scheduleAt(simTime() + sendDelay, scheduleMsg);
}

void RadioPacketGenerator::sendPacket() {
    // Time to send a packet!!
    inet::ieee80211::Ieee80211Frame *newPkt = new inet::ieee80211::Ieee80211Frame("RadioPacketGeneratorPkt");
    // set the length of the packet
    newPkt->setByteLength(generatedPktLength);
    newPkt->setDuration(0);

    send(newPkt, "radioOut");

    EV << "[JAMMER]: Delivered a packet to the radio for transmission\n";
}

void RadioPacketGenerator::finish() {
    cancelAndDelete(scheduleMsg);
}
